use std::str::FromStr;

use palette::{Darken, FromColor, Lighten, Mix, Oklch, Srgb};
use wasm_bindgen::prelude::wasm_bindgen;

use crate::errors;

#[wasm_bindgen]
pub fn lighten(color: &str, percent: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let lightened_color = oklch.lighten(percent);
    let srgb = Srgb::from_color(lightened_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn lighten_absolute(color: &str, value: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let lightened_color = oklch.lighten_fixed(value);
    let srgb = Srgb::from_color(lightened_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn darken(color: &str, percent: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let darkened_color = oklch.darken(percent);
    let srgb = Srgb::from_color(darkened_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn darken_absolute(color: &str, value: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let oklch = Oklch::from_color(origin_color);
    let darkened_color = oklch.darken_fixed(value);
    let srgb = Srgb::from_color(darkened_color);
    Ok(format!("{:x}", srgb.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn mix(color1: &str, color2: &str, percent: f32) -> Result<String, errors::ColorError> {
    let srgb1 = Srgb::from_str(color1)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color1.to_string()))?
        .into_format::<f32>();
    let srgb2 = Srgb::from_str(color2)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color2.to_string()))?
        .into_format::<f32>();
    let mixed_color = srgb1.mix(srgb2, percent);
    Ok(format!("{:x}", mixed_color.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn tint(color: &str, percent: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let tinted_color = origin_color.mix(Srgb::new(1.0, 1.0, 1.0), percent);
    Ok(format!("{:x}", tinted_color.into_format::<u8>()))
}

#[wasm_bindgen]
pub fn shade(color: &str, percent: f32) -> Result<String, errors::ColorError> {
    let origin_color = Srgb::from_str(color)
        .map_err(|_| errors::ColorError::UnrecogniazedRGB(color.to_string()))?
        .into_format::<f32>();
    let shaded_color = origin_color.mix(Srgb::new(0.0, 0.0, 0.0), percent);
    Ok(format!("{:x}", shaded_color.into_format::<u8>()))
}
