use std::collections::HashMap;

use serde::{ser::SerializeStruct, Serialize};

use crate::{convert::map_hsl_to_srgb_hex, errors, schemes::material_design_2::swatch::M2Swatch};

use super::{color_set::M2ColorSet, swatch::SwatchIndex};

#[derive(Debug, Clone)]
pub struct M2BaselineColors {
    pub primary: M2ColorSet,
    pub secondary: M2ColorSet,
    pub error: M2ColorSet,
    pub background: M2ColorSet,
    pub surface: M2ColorSet,
    pub shadow: String,
    pub custom_colors: HashMap<String, M2ColorSet>,
    neutral_swatch: M2Swatch,
    dark_set: bool,
}

impl M2BaselineColors {
    pub fn new(
        primary_color: &str,
        secondary_color: &str,
        error_color: &str,
        dark_baseline: bool,
    ) -> Result<Self, errors::ColorError> {
        let primary_swatch = M2Swatch::from_color(primary_color)?;
        let secondary_swatch = M2Swatch::from_color(secondary_color)?;
        let error_swatch = M2Swatch::from_color(error_color)?;
        let neutral_swatch = M2Swatch::default_neutral();

        Ok(Self {
            primary: M2ColorSet::from_swatch(
                &primary_swatch,
                &neutral_swatch,
                dark_baseline,
                None,
            )?,
            secondary: M2ColorSet::from_swatch(
                &secondary_swatch,
                &neutral_swatch,
                dark_baseline,
                None,
            )?,
            error: M2ColorSet::from_swatch(&error_swatch, &neutral_swatch, dark_baseline, None)?,
            background: M2ColorSet::surface(&neutral_swatch, dark_baseline)?,
            surface: M2ColorSet::surface(&neutral_swatch, dark_baseline)?,
            shadow: map_hsl_to_srgb_hex(&neutral_swatch.tone(SwatchIndex::SI900)),
            custom_colors: HashMap::new(),
            neutral_swatch,
            dark_set: dark_baseline,
        })
    }

    pub fn add_custom_color(&mut self, name: &str, color: &str) -> Result<(), errors::ColorError> {
        let swatch = M2Swatch::from_color(color)?;
        self.custom_colors.insert(
            name.to_string(),
            M2ColorSet::from_swatch(&swatch, &self.neutral_swatch, self.dark_set, None)?,
        );
        Ok(())
    }

    pub fn to_css_variable(&self) -> Vec<String> {
        let mut variable_lines = Vec::new();
        let prefix = if self.dark_set { "dark" } else { "light" };

        variable_lines.extend(self.primary.to_css_variable(&prefix, "primary"));
        variable_lines.extend(self.secondary.to_css_variable(&prefix, "secondary"));
        variable_lines.extend(self.error.to_css_variable(&prefix, "error"));
        variable_lines.extend(self.background.to_css_variable(&prefix, "background"));
        variable_lines.extend(self.surface.to_css_variable(&prefix, "surface"));
        variable_lines.push(format!("--color-{}-shadow: #{};", prefix, self.shadow));

        for (name, color_set) in &self.custom_colors {
            variable_lines.extend(color_set.to_css_variable(&prefix, &name.to_lowercase()));
        }

        variable_lines
    }

    pub fn to_scss_variable(&self) -> Vec<String> {
        let mut variable_lines = Vec::new();
        let prefix = if self.dark_set { "dark" } else { "light" };

        variable_lines.extend(self.primary.to_scss_variable(&prefix, "primary"));
        variable_lines.extend(self.secondary.to_scss_variable(&prefix, "secondary"));
        variable_lines.extend(self.error.to_scss_variable(&prefix, "error"));
        variable_lines.extend(self.background.to_scss_variable(&prefix, "background"));
        variable_lines.extend(self.surface.to_scss_variable(&prefix, "surface"));
        variable_lines.push(format!("$color-{}-shadow: #{};", prefix, self.shadow));

        for (name, color_set) in &self.custom_colors {
            variable_lines.extend(color_set.to_scss_variable(&prefix, &name.to_lowercase()));
        }

        variable_lines
    }

    pub fn to_javascript_object(&self) -> Vec<String> {
        let mut variable_lines = Vec::new();
        let prefix = if self.dark_set { "dark" } else { "light" };

        variable_lines.extend(self.primary.to_javascript_object(&prefix, "primary"));
        variable_lines.extend(self.secondary.to_javascript_object(&prefix, "secondary"));
        variable_lines.extend(self.error.to_javascript_object(&prefix, "error"));
        variable_lines.extend(self.background.to_javascript_object(&prefix, "background"));
        variable_lines.extend(self.surface.to_javascript_object(&prefix, "surface"));
        variable_lines.push(format!("{}Shadow: '#{}',", prefix, self.shadow));

        for (name, color_set) in &self.custom_colors {
            variable_lines.extend(color_set.to_javascript_object(&prefix, &name.to_lowercase()));
        }

        variable_lines
    }
}

impl Serialize for M2BaselineColors {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer,
    {
        let mut map = serializer.serialize_struct("baseline", 7)?;
        map.serialize_field("primary", &self.primary)?;
        map.serialize_field("secondary", &self.secondary)?;
        map.serialize_field("error", &self.error)?;
        map.serialize_field("background", &self.background)?;
        map.serialize_field("surface", &self.surface)?;
        map.serialize_field("shadow", &self.shadow)?;
        map.serialize_field("custom_colors", &self.custom_colors)?;
        map.end()
    }
}
