use serde::Serialize;

use crate::convert::map_lch_to_srgb_hex;

use super::tonal_palette::TonalPalette;

#[derive(Debug, Clone, Serialize)]
pub struct M3SurfaceSet {
    pub root: String,
    pub dim: String,
    pub bright: String,
    pub variant: String,
    pub container: String,
    pub container_lowest: String,
    pub container_low: String,
    pub container_high: String,
    pub container_highest: String,
    pub on_root: String,
    pub on_root_variant: String,
    pub inverse: String,
    pub on_inverse: String,
}

impl M3SurfaceSet {
    pub fn new_light_set(neutral: &TonalPalette, neutral_variant: &TonalPalette) -> Self {
        let root = neutral.tone(98.0);
        let dim = neutral.tone(87.0);
        let bright = neutral.tone(98.0);
        let variant = neutral_variant.tone(90.0);
        let container = neutral.tone(94.0);
        let container_lowest = neutral.tone(100.0);
        let container_low = neutral.tone(96.0);
        let container_high = neutral.tone(92.0);
        let container_highest = neutral.tone(90.0);
        let on_root = neutral_variant.tone(10.0);
        let on_root_variant = neutral_variant.tone(30.0);
        let inverse = neutral.tone(20.0);
        let on_inverse = neutral_variant.tone(95.0);

        Self {
            root: map_lch_to_srgb_hex(&root),
            dim: map_lch_to_srgb_hex(&dim),
            bright: map_lch_to_srgb_hex(&bright),
            variant: map_lch_to_srgb_hex(&variant),
            container: map_lch_to_srgb_hex(&container),
            container_lowest: map_lch_to_srgb_hex(&container_lowest),
            container_low: map_lch_to_srgb_hex(&container_low),
            container_high: map_lch_to_srgb_hex(&container_high),
            container_highest: map_lch_to_srgb_hex(&container_highest),
            on_root: map_lch_to_srgb_hex(&on_root),
            on_root_variant: map_lch_to_srgb_hex(&on_root_variant),
            inverse: map_lch_to_srgb_hex(&inverse),
            on_inverse: map_lch_to_srgb_hex(&on_inverse),
        }
    }

    pub fn new_dark_set(neutral: &TonalPalette, neutral_variant: &TonalPalette) -> Self {
        let root = neutral.tone(6.0);
        let dim = neutral.tone(6.0);
        let bright = neutral.tone(24.0);
        let variant = neutral_variant.tone(30.0);
        let container = neutral.tone(12.0);
        let container_lowest = neutral.tone(4.0);
        let container_low = neutral.tone(10.0);
        let container_high = neutral.tone(17.0);
        let container_highest = neutral.tone(22.0);
        let on_root = neutral_variant.tone(90.0);
        let on_root_variant = neutral_variant.tone(80.0);
        let inverse = neutral.tone(90.0);
        let on_inverse = neutral_variant.tone(20.0);

        Self {
            root: map_lch_to_srgb_hex(&root),
            dim: map_lch_to_srgb_hex(&dim),
            bright: map_lch_to_srgb_hex(&bright),
            variant: map_lch_to_srgb_hex(&variant),
            container: map_lch_to_srgb_hex(&container),
            container_lowest: map_lch_to_srgb_hex(&container_lowest),
            container_low: map_lch_to_srgb_hex(&container_low),
            container_high: map_lch_to_srgb_hex(&container_high),
            container_highest: map_lch_to_srgb_hex(&container_highest),
            on_root: map_lch_to_srgb_hex(&on_root),
            on_root_variant: map_lch_to_srgb_hex(&on_root_variant),
            inverse: map_lch_to_srgb_hex(&inverse),
            on_inverse: map_lch_to_srgb_hex(&on_inverse),
        }
    }

    pub fn to_css_variables(&self, prefix: &str) -> Vec<String> {
        let mut css_variables = Vec::new();

        css_variables.push(format!("--color-{}-surface: #{};", prefix, self.root));
        css_variables.push(format!("--color-{}-surface-dim: #{};", prefix, self.dim));
        css_variables.push(format!(
            "--color-{}-surface-bright: #{};",
            prefix, self.bright
        ));
        css_variables.push(format!(
            "--color-{}-surface-variant: #{};",
            prefix, self.variant
        ));
        css_variables.push(format!(
            "--color-{}-surface-container: #{};",
            prefix, self.container
        ));
        css_variables.push(format!(
            "--color-{}-surface-container-lowest: #{};",
            prefix, self.container_lowest
        ));
        css_variables.push(format!(
            "--color-{}-surface-container-low: #{};",
            prefix, self.container_low
        ));
        css_variables.push(format!(
            "--color-{}-surface-container-high: #{};",
            prefix, self.container_high
        ));
        css_variables.push(format!(
            "--color-{}-surface-container-highest: #{};",
            prefix, self.container_highest
        ));
        css_variables.push(format!("--color-{}-on-surface: #{};", prefix, self.on_root));
        css_variables.push(format!(
            "--color-{}-on-surface-variant: #{};",
            prefix, self.on_root_variant
        ));
        css_variables.push(format!(
            "--color-{}-inverse-surface: #{};",
            prefix, self.inverse
        ));
        css_variables.push(format!(
            "--color-{}-inverse-on-surface: #{};",
            prefix, self.on_inverse
        ));

        css_variables
    }

    pub fn to_scss_variables(&self, prefix: &str) -> Vec<String> {
        let mut scss_variables = Vec::new();

        scss_variables.push(format!("$color-{}-surface: #{};", prefix, self.root));
        scss_variables.push(format!("$color-{}-surface-dim: #{};", prefix, self.dim));
        scss_variables.push(format!(
            "$color-{}-surface-bright: #{};",
            prefix, self.bright
        ));
        scss_variables.push(format!(
            "$color-{}-surface-variant: #{};",
            prefix, self.variant
        ));
        scss_variables.push(format!(
            "$color-{}-surface-container: #{};",
            prefix, self.container
        ));
        scss_variables.push(format!(
            "$color-{}-surface-container-lowest: #{};",
            prefix, self.container_lowest
        ));
        scss_variables.push(format!(
            "$color-{}-surface-container-low: #{};",
            prefix, self.container_low
        ));
        scss_variables.push(format!(
            "$color-{}-surface-container-high: #{};",
            prefix, self.container_high
        ));
        scss_variables.push(format!(
            "$color-{}-surface-container-highest: #{};",
            prefix, self.container_highest
        ));
        scss_variables.push(format!("$color-{}-on-surface: #{};", prefix, self.on_root));
        scss_variables.push(format!(
            "$color-{}-on-surface-variant: #{};",
            prefix, self.on_root_variant
        ));
        scss_variables.push(format!(
            "$color-{}-inverse-surface: #{};",
            prefix, self.inverse
        ));
        scss_variables.push(format!(
            "$color-{}-inverse-on-surface: #{};",
            prefix, self.on_inverse
        ));

        scss_variables
    }

    pub fn to_javascript_object_fields(&self, prefix: &str) -> Vec<String> {
        let mut js_object_fields = Vec::new();

        js_object_fields.push(format!("{}Surface: '#{}',", prefix, self.root));
        js_object_fields.push(format!("{}SurfaceDim: '#{}',", prefix, self.dim));
        js_object_fields.push(format!("{}SurfaceBright: '#{}',", prefix, self.bright));
        js_object_fields.push(format!("{}SurfaceVariant: '#{}',", prefix, self.variant));
        js_object_fields.push(format!(
            "{}SurfaceContainer: '#{}',",
            prefix, self.container
        ));
        js_object_fields.push(format!(
            "{}SurfaceContainerLowest: '#{}',",
            prefix, self.container_lowest
        ));
        js_object_fields.push(format!(
            "{}SurfaceContainerLow: '#{}',",
            prefix, self.container_low
        ));
        js_object_fields.push(format!(
            "{}SurfaceContainerHigh: '#{}',",
            prefix, self.container_high
        ));
        js_object_fields.push(format!(
            "{}SurfaceContainerHighest: '#{}',",
            prefix, self.container_highest
        ));
        js_object_fields.push(format!("{}OnSurface: '#{}',", prefix, self.on_root));
        js_object_fields.push(format!(
            "{}OnSurfaceVariant: '#{}',",
            prefix, self.on_root_variant
        ));
        js_object_fields.push(format!("{}InverseSurface: '#{}',", prefix, self.inverse));
        js_object_fields.push(format!(
            "{}InverseOnSurface: '#{}',",
            prefix, self.on_inverse
        ));

        js_object_fields
    }
}
