import { useMemo } from 'react';
import { useColorFunction } from '../../../ColorFunctionContext';
import { ScrollArea } from '../../../components/ScrollArea';
import { Baseline, ColorSet, MaterialDesign2SchemeStorage } from '../../../material-2-scheme';
import { SchemeContent } from '../../../models';
import styles from './Preview.module.css';

type M2SchemeFunctionColorProps = {
  title: string;
  color: ColorSet;
};

function M2SchemeFunctionColor({ title, color }: M2SchemeFunctionColorProps) {
  const { colorFn } = useColorFunction();
  const rootWacgRatio = useMemo(() => {
    try {
      return colorFn?.wacg_relative_contrast(color.on, color.root) ?? null;
    } catch (e) {
      console.error('[calc root wacg]', e);
    }
    return null;
  }, [colorFn, color.on, color.root]);
  const variantWacgRatio = useMemo(() => {
    try {
      return colorFn?.wacg_relative_contrast(color.on, color.variant) ?? null;
    } catch (e) {
      console.error('[calc variant wacg]', e);
    }
    return null;
  }, [colorFn, color.on, color.variant]);

  return (
    <div className={styles.horizontal_set}>
      <div
        className={styles.color_block}
        style={{ backgroundColor: `#${color.root}`, color: `#${color.on}` }}>
        <span>{title}</span>
        <span className={styles.wacg}>WACG: {rootWacgRatio?.toFixed(2)}</span>
      </div>
      <div
        className={styles.color_block}
        style={{ backgroundColor: `#${color.variant}`, color: `#${color.on}` }}>
        <span>{title} Variant</span>
        <span className={styles.wacg}>WACG: {variantWacgRatio?.toFixed(2)}</span>
      </div>
      <div
        className={styles.color_block}
        style={{ backgroundColor: `#${color.on}`, color: `#${color.root}` }}>
        On {title}
      </div>
    </div>
  );
}

type PreviewBlockProps = {
  title: string;
  baseline: Baseline;
};

function PreviewBlock({ title, baseline }: PreviewBlockProps) {
  return (
    <div
      className={styles.preview_block}
      style={{
        backgroundColor: `#${baseline.background.root}`,
        color: `#${baseline.background.on}`,
      }}>
      <h4>{title} Scheme</h4>
      <M2SchemeFunctionColor title="Primary" color={baseline.primary} />
      <M2SchemeFunctionColor title="Secondary" color={baseline.secondary} />
      <M2SchemeFunctionColor title="Error" color={baseline.error} />
      <M2SchemeFunctionColor title="Background" color={baseline.background} />
      <M2SchemeFunctionColor title="Surface" color={baseline.surface} />
      {Object.entries(baseline.custom_colors).map(([name, set], index) => (
        <M2SchemeFunctionColor key={index} title={name} color={set} />
      ))}
    </div>
  );
}

type M2SchemePreviewProps = {
  scheme: SchemeContent<MaterialDesign2SchemeStorage>;
};

export function M2SchemePreview({ scheme }: M2SchemePreviewProps) {
  return (
    <ScrollArea enableY>
      <div className={styles.preview_layout}>
        <PreviewBlock title="Light" baseline={scheme.schemeStorage.scheme!.light} />
        <PreviewBlock title="Dark" baseline={scheme.schemeStorage.scheme!.dark} />
      </div>
    </ScrollArea>
  );
}
